# $Id: emupload.py,v 1.7 2011/05/04 08:15:25 irees Exp $
import Queue
import collections
import datetime
import functools
import getpass
import glob
import operator
import optparse
import os
import re
import sys
import time

# PyQt4 imports
from PyQt4 import QtGui, QtCore, Qt

# emdash imports
import emdash.config
import emdash.emmodels
import emdash.emwizard
import emdash.emthreads
import emdash.ui



def error(e):
	print "Showing error:", e
	error = QtGui.QMessageBox()
	error.setWindowTitle("Error")
	error.setText(unicode(e))
	error.setIcon(QtGui.QMessageBox.Critical)
	error.exec_()





def main():
	parser = emdash.config.DBRPCOptions()
	usage = """%prog"""
	parser.set_usage(usage)	

	parser.add_option("--microscope", "-m", type="int",  help="Microscope or Microscope folder Record name (e.g. 110, 114..)")
	parser.add_option("--session_rectype","-s", type="string", help="Session Protocol (Default is microscopy)", default='microscopy')

	group = optparse.OptionGroup(parser, "Scan Conversion Options")
	group.add_option("--tif2mrc", metavar="<0|1>", dest="tif2mrc", type="int",  help="Scan Option: See nikontiff2mrc for options and defaults for the following options:", default=True)		
	group.add_option("--bin", metavar="<n>", dest="bin", type="float", help="Scan Option: bin the pixels. default to 1 (no binning)")	
	group.add_option("--invert", metavar="<0|1>", dest="invert", type="int", help="Scan Option: invert the contrast to make particles white. default to 0")	
	group.add_option("--ODconversion", metavar="<0|1>", dest="odconversion", type="int",  help="Scan Option: if convert pixel value from transmitted light intensity to O.D. of the film. enabled by default")
	parser.add_option_group(group)

	(options, args) = parser.parse_args()
	db = parser.opendb()

	emdash.config.set('microscope', options.microscope)
	emdash.config.set('session_rectype', options.session_rectype)

	app = QtGui.QApplication(sys.argv)
	window = MicroscopeUpload(db=db)
	window.login()
	sys.exit(app.exec_())	




class EMDashController(QtCore.QObject):		

	# Update the record mapping, cached records, and recnames.
	signal_update_recs = QtCore.pyqtSignal(object, object, object)
	signal_update_comments = QtCore.pyqtSignal(object)

	# Changes in file status
	signal_update_setting = QtCore.pyqtSignal(unicode, unicode)
	signal_newfile = QtCore.pyqtSignal(unicode, object)
	signal_file_status = QtCore.pyqtSignal(unicode, object)
	signal_file_success = QtCore.pyqtSignal(unicode, object)
	signal_file_failure = QtCore.pyqtSignal(unicode, unicode)

	# Session controls
	signal_begin_session = QtCore.pyqtSignal(int)
	signal_end_session = QtCore.pyqtSignal()
	signal_newgrid = QtCore.pyqtSignal(int)

	# Login
	signal_getlogin = QtCore.pyqtSignal(unicode, unicode)
	signal_login = QtCore.pyqtSignal(unicode, unicode, unicode)
	signal_auth_exception = QtCore.pyqtSignal(unicode)
	signal_exception = QtCore.pyqtSignal(unicode)

	# Change a value in a record: name, parameter, new value
	signal_set = QtCore.pyqtSignal(int, unicode, unicode)

	# Change a controller setting: parameter, new value
	signal_setting = QtCore.pyqtSignal(unicode, object)


	def __init__(self, args=None, options=None, parent=None, db=None):
		super(EMDashController, self).__init__(parent=parent)

		# Record IDs, record names, comments
		self.db = db
		self.bdos = {}
		self.comments = []
		self.recs = {}
		self.recnames = {}		
		self.names = {}

		# Current microscope settings...		
		self.settings = {}
		self.settings['assess_image_quality'] = 5

		# Queue of currently un-uploaded files
		self.queue = Queue.Queue()

		self._setupworkers()



	###############################
	# Other Threads
	###############################

	def _setupworkers(self):
		# Starts when a new microscopy session is made
		self.clockthread = emdash.emthreads.ClockThread(parent=self)

		# FSPollThread starts whenever a directory is selected
		self.fspollthread = emdash.emthreads.FSPollThread(parent=self) # queue=self.queue
		self.fspollthread.signal_newfile.connect(self.newfile, type=QtCore.Qt.QueuedConnection)

		# Starts when a grid has been selected or created
		self.uploadthread = emdash.emthreads.UploadThread(parent=self, db=self.db, queue=self.queue)

		self.uploadthread.signal_file_status.connect(self.signal_file_status, type=QtCore.Qt.QueuedConnection)
		self.uploadthread.signal_file_success.connect(self.signal_file_success, type=QtCore.Qt.QueuedConnection)
		self.uploadthread.signal_file_failure.connect(self.signal_file_failure, type=QtCore.Qt.QueuedConnection)

		# Gracefully stop processing when session ends..
		self.signal_end_session.connect(self.fspollthread.quit, type=QtCore.Qt.QueuedConnection)
		self.signal_end_session.connect(self.clockthread.quit, type=QtCore.Qt.QueuedConnection)



	###############################
	# Session/Record ID management
	###############################

	def _getnames(self):
		return filter(None, self.names.values())


	def _reset(self, exclude=None):
		"""Reset the controller"""

		if exclude == None:
			exclude = []
		if not hasattr(exclude, "__iter__"):
			exclude = [exclude]

		exclude.append('microscope')
		backup = {}

		for i in exclude:
			if self.names.get(i) != None:
				backup[i] = self.names.get(i)

		self.names = backup
		self.update_recs()


	def run(self):
		try:
			self.db.logout()
		except:
			pass
		# print "Waiting for queue to empty..."
		# self.queue.put({"filename":None})
		# self.uploadthread.wait()
		# self.finished.emit()
		# self.db.logout()
		# print "Queue is empty; closing session..."



	@QtCore.pyqtSlot(str, str)
	def login(self, name=None, password=None):		
		self._reset()
		self.db._setContext(None)

		try:
			self.db.login(name, password)
			name, groups = self.db.checkcontext()
			user = self.db.getuser(name) or {}
			displayname = user.get('displayname', name)
			self.signal_login.emit(self.db._ctx.ctxid, name, displayname)						

		except Exception, e:
			self.signal_auth_exception.emit(unicode(e))
			return


	@QtCore.pyqtSlot(unicode, unicode, unicode)
	def begin_no_session(self, ctxid, name, displayname):	
		# begin_session directly from login callback
		self.clockthread.start()		
		self.uploadthread.start()
		self.update_comments()
		self.update_recs()
		self.signal_begin_session.emit(-1)


	@QtCore.pyqtSlot(object, object)
	def begin_session(self, editrecords, select_records):	
		# Find the correct microscope
		self.clockthread.start()		
		self.uploadthread.start()
		self.update_comments()
		self.update_recs()
		self.signal_begin_session.emit(self.names.get(emdash.config.get('session_rectype')))


	@QtCore.pyqtSlot()	
	def end_session(self, restart=True):
		self._reset()
		self.signal_end_session.emit()		
		# self.clockthread.quit()
		# self.fspollthread.quit()
		# self.start()


	# Controller discovers a new file, does some processing, and rebroadcasts.	
	# @QtCore.pyqtSlot(unicode, object)	
	def newfile(self, filename, d=None):
		d = d or {}
		filename = unicode(filename)

		ext = os.path.splitext(filename)[-1]
		auto = False
		if ext in emdash.config.get('autoqueue'):
			auto = True

		if ext not in emdash.config.get('autoqueue') + emdash.config.get('manualqueue'):
			return

		d.update({'_recname': os.path.basename(filename), '_filename':filename, "_auto":auto})
		d.update(self.settings)
		self.signal_newfile.emit(filename, d)



	###############################
	# Editing records / Comments
	###############################

	@QtCore.pyqtSlot(int, unicode, unicode)
	def set(self, name, param, value):
		if name < 0:
			self.signal_exception.emit("Cannot update non-existent record")
			return

		try:
			rec = self.db.putrecordvalues(name, {unicode(param): unicode(value)})
			self.signal_file_status.emit(str(name), rec)
		except Exception, inst:
			self.signal_exception.emit("Cannot update record: %s"%inst)


	@QtCore.pyqtSlot(int, unicode)
	def addcomment(self, name, text):
		try:
			self.db.addcomment(name, unicode(text))
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			return

		self.update_comments()


	def _getcomments(self):
		return []


	def update_comments(self):
		self.comments = self._getcomments()
		self.signal_update_comments.emit(self.comments)


	def update_recs(self, newmapping=None, quiet=False):
		"""Update the current set of records."""

		newmapping = newmapping or {}
		if newmapping:
			self.names.update(newmapping)

		rnget = self._getnames()

		recs = self.db.getrecord(rnget)
		for rec in recs:
			self.recs[rec.get('name')] = rec

		rn = self.db.renderview(rnget, viewtype="recname")
		self.recnames.update(rn)		

		if not quiet:
			self.signal_update_recs.emit(self.names, self.recs, self.recnames)


	def set_setting(self, param, value):
		# print "Changing setting: %s -> %s"%(param, value)
		self.settings[str(param)] = str(value)
		self.signal_update_setting.emit(str(param), str(value))



	###############################
	# Some controller specific behaviors...
	###############################

	# Fold this into something else..
	def newmicrograph(self):
		id_micrograph = self.settings.get('id_micrograph','')
		matches = re.search('(\w+?)(\d+)', id_micrograph)
		if not matches:
			print "Error creating new micrograph: No Micrograph ID..."
			return

		matches = matches.groups()		
		newid = int(matches[1])+1
		newid = matches[0] + str(newid).zfill(len(matches[1]))

		data = {'_recname': id_micrograph, '_filename':id_micrograph, 'id_micrograph':id_micrograph, "_auto":True, 'rectype':'micrograph'}
		data.update(self.settings)

		# Emit the new file signal, and update the UI to show the new file micrograph ID
		self.signal_newfile.emit(id_micrograph, data)
		self.set_setting('id_micrograph', newid)


	@QtCore.pyqtSlot(object, object)
	def newfreezing(self, newrecords, selected_records):
		try:
			newrecords = self.db.putrecord(newrecords)
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			raise


	@QtCore.pyqtSlot(int)
	def selectmicroscopy(self, name):
		self.begin_session([], {emdash.config.get('session_rectype'):name})


	@QtCore.pyqtSlot(int)
	def selectgrid(self, name):
		self.newgrid([], {'grid_imaging':name})


	# ian: short-circuit signal since everything is converted to QStrings and blows up xmlrpc
	@QtCore.pyqtSlot(object, object)
	def newgrid(self, newrecords, selected_records):
		self._reset(exclude = emdash.config.get('session_rectype'))

		try:
			newrecords = self.db.putrecord(newrecords)
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			raise

		for rec in newrecords:
			selected_records[rec.get("rectype")] = rec.get("name")

		self.update_recs(newmapping=selected_records)		

		# ian: todo. I changed to use page names, instead of record types.. but this may cause some problems.
		e = selected_records.get("grid_imaging")
		if e > -1:
			self.signal_newgrid.emit(int(selected_records.get("grid_imaging")))


	@QtCore.pyqtSlot(object, object)
	def edit_records(self, editrecords, selected_records):
		recs = {}
		# xmlrpc requires keys to be strs
		for rec in editrecords:
			recs[str(rec['name'])] = rec

		try:
			self.db.putrecord(recs.values())
		except Exception, e:
			self.signal_exception.emit(unicode(e))
			raise

		self.update_recs()




class MicroscopeController(EMDashController):

	def _getcomments(self):
		root = self.names.get('microscope')
		if  root != None:
			recs = self.db.getchildren(self.names["microscope"], 1, ["microscopy*"])
			coms = self.db.getcomments(self._getnames()+recs)			
			# update the recnames..
			self.recnames.update(self.db.renderview([i[0] for i in coms], viewtype="recname"))
			return coms
			# ret = collections.defaultdict(list) # backwards compat
			# for i in coms:
			# 	ret[i[0]].append(i[1:])
			# return ret



	def begin_session(self, editrecords, select_records):		
		# Find the correct microscope
		t = self.db.gettime()

		if emdash.config.get('microscope') == None:
			raise ValueError, "No microscope specified!"

		childrecs = set(self.db.getchildren(emdash.config.get('microscope'), 1, "microscope"))
		childrecs.add(emdash.config.get('microscope'))
		microscopes = self.db.getrecord(list(childrecs))
		microscope = None

		for m in microscopes:
			# print m.get('name'), m.get('date_start'), m.get('date_end')
			if m.get('date_start') <= t < m.get('date_end'):
				self.names['microscope'] = m.get('name')

		if not self.names.get('microscope'):
			raise ValueError, "No microscope valid for today's date: %s... tried children of %s"%(t, emdash.config.get('microscope'))

		if select_records.get(emdash.config.get('session_rectype')):
			rec = self.db.getrecord(select_records.get(emdash.config.get('session_rectype')))

		else:
			rec = self.db.newrecord(emdash.config.get('session_rectype'), self.names.get("microscope"))
			rec["date_start"] = self.db.gettime()
			rec["parents"] = [self.names.get("microscope")]
			rec.update(editrecords[0])
			rec = self.db.putrecord(rec)

		self.names[emdash.config.get('session_rectype')] = rec.get("name")
		self.clockthread.start()		
		self.uploadthread.start()

		self.update_comments()
		self.update_recs()
		self.signal_begin_session.emit(self.names[emdash.config.get('session_rectype')])



	def end_session(self, restart=True):
		if self.names.get(emdash.config.get('session_rectype')) != None:
			try:
				rec = self.db.getrecord(self.names.get(emdash.config.get('session_rectype')))
				rec["date_end"] = self.db.gettime()
				rec = self.db.putrecord(rec)
			except:
				print "Error ending session record"

		self._reset()
		self.signal_end_session.emit()
		# self.clockthread.quit()
		# self.fspollthread.quit()
		# self.start()







# ian: todo: @abstractmethod decorators in BaseUpload
class BaseUpload(QtGui.QMainWindow):
	# Base class for an Upload UI

	signal_set = QtCore.pyqtSignal(int, unicode, unicode)
	signal_set_setting = QtCore.pyqtSignal(unicode, object)

	def __init__(self, parent=None, db=None):
		super(BaseUpload, self).__init__(parent=parent)
		
		# Get a DBProxy handle
		self.db = db
		
		# Items to display, usually set from controller update signals..
		self.name = None
		self.displayname = None
		self.names = {}
		self.recnames = {}
		self.comments = []
		self.path = None		
		self.end = False		
				
		# Setup Controller
		self.controller = self.get_controller()

		# Display model for the queue
		headers = ["_recname", "tem_magnification_set", "ctf_defocus_set", 'tem_dose_rate', 'time_exposure_tem', "assess_image_quality", "_status"]
		headernames = ["Name", "Mag", "Defocus", "Dose", "Exposure", "Quality", "Status"]
		self.queuemodel = emdash.emmodels.QueueModel(headers=headers, headernames=headernames, queue=self.controller.queue)

		# Bind controller signals
		self.controller.signal_newgrid.connect(self.queuemodel.newgrid, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_newgrid.connect(self.newgrid, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_update_recs.connect(self.update_recs, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_update_comments.connect(self.update_comments, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_update_setting.connect(self.update_setting, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_begin_session.connect(self.begin_session, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_exception.connect(self.error, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_login.connect(self.login_cb, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_auth_exception.connect(self.login, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_file_status.connect(self.queuemodel.file_status, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_file_success.connect(self.queuemodel.file_success, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_file_failure.connect(self.queuemodel.file_failure, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_end_session.connect(self.end_session, type=QtCore.Qt.QueuedConnection)

		# Clock		
		self.controller.clockthread.signal_tick.connect(self.tock, type=QtCore.Qt.QueuedConnection)

		# Update a single item in a record
		self.signal_set.connect(self.controller.set, type=QtCore.Qt.QueuedConnection)
		self.signal_set_setting.connect(self.controller.set_setting, type=QtCore.Qt.QueuedConnection)

		# Setup UI and other slots
		self.ui = self.get_ui()

		# Some basic signals that are common across most Uploaders..
		self.init_common()
		
		# Subclass init
		self.init()


	def init(self):
		# The default Uploader doesn't have any special session management
		self.controller.signal_login.connect(self.controller.begin_no_session, type=QtCore.Qt.QueuedConnection)
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session", self, triggered=self._select_grid_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session by Record ID", self, triggered=self._name_grid_wizard))
						
			
	def init_common(self):
		# Connect the controller's queue model to self.ui.tree_files
		self.ui.tree_files.setModel(self.queuemodel)
		self.ui.tree_files.setColumnWidth(0, 250)
		self.ui.tree_files.setColumnWidth(1, 60)
		self.ui.tree_files.setColumnWidth(2, 60)
		self.ui.tree_files.setColumnWidth(3, 60)
		self.ui.tree_files.setColumnWidth(4, 60)
		self.ui.tree_files.setColumnWidth(5, 80)
		self.ui.tree_files.setColumnWidth(6, 50)

		# Connect all the newfile events, so the file widget updates
		self.controller.signal_newfile.connect(self.ui.tree_files.newfile, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_newfile.connect(self.queuemodel.newfile, type=QtCore.Qt.QueuedConnection)

		# Connect UI signals back to controller
		self.ui.tree_files.signal_set.connect(self.controller.set, type=QtCore.Qt.QueuedConnection)
		self.ui.tree_files.signal_enqueue.connect(self.controller.uploadthread.enqueue, type=QtCore.Qt.QueuedConnection)
		
		# Drop-down buttons
		self.ui.button_path.addAction(QtGui.QAction("Select directory", self, triggered=self.setpath))

		self.ui.button_session.addAction(QtGui.QAction("Logout", self, triggered=self.controller.end_session))

		self.connect_edits()


	def connect_edits(self):
		pass
		

	def get_controller(self):
		return EMDashController(db=emdash.config.copydb(self.db))


	def get_ui(self):
		u = emdash.ui.Ui_Upload.Ui_Upload()
		u.setupUi(self)
		return u


	@QtCore.pyqtSlot(str)
	def error(self, e):
		error(e)




	##############################
	# Set / update settings
	##############################

	def set_setting(self, value, param):
		# Change a setting, usually sends to controller
		# print "upload: set_setting: %s %s"%(value, param)
		if param:
			self.signal_set_setting.emit(param, value)
			

	def update_ui(self, comments=False):
		# Update the UI from a Record change in the controller.
		self.ui.label_grid.setText(self.getlabel(["grid_imaging"]))


	def update_setting(self, param, value):	
		# Update a UI setting, usually from controller
		pass
			

	@QtCore.pyqtSlot(object, object, object)
	def update_recs(self, names, recs, recnames):
		# Controller saved some records, callback to the UI to reflect this
		self.names = names
		self.recs = recs
		self.recnames = recnames
		self.update_ui()
				
	
	@QtCore.pyqtSlot(object)
	def update_comments(self, l):
		# Callback for updated comments
		self.comments = l
		self.update_ui(comments=True)
		

	@QtCore.pyqtSlot(str)
	def update_path(self, path):
		self.ui.label_path.setText(unicode(path))


	@QtCore.pyqtSlot(int)
	def newgrid(self, newgrid):
		if newgrid > -1:
			self.ui.button_path.setEnabled(True)
			self.ui.frame_settings.setEnabled(True)
			try: self.ui.button_newmicrograph.setEnabled(True)
			except: pass
		else:
			self.ui.button_path.setEnabled(False)
			self.ui.frame_settings.setEnabled(False)
			try: self.ui.button_newmicrograph.setEnabled(False)
			except: pass



	##############################
	# Session
	##############################

	# @QtCore.pyqtSlot()
	@QtCore.pyqtSlot(unicode)
	@QtCore.pyqtSlot(unicode, unicode, unicode)
	def login(self, authmsg=None, name=None, password=None, microscope=None):
		# Get login info for new session
		name = name or emdash.config.get('name')
		password = password or emdash.config.get('password')		
		microscope = microscope or emdash.config.get('microscope')

		l = Login(parent=self, authmsg=authmsg, name=name, password=password, microscope=microscope)
		l.signal_getlogin.connect(self.controller.login)
		if name and password:
			l.signal_getlogin.emit(name, password)
		else:
			l.show()


	@QtCore.pyqtSlot(unicode, unicode, unicode)
	def login_cb(self, ctxid, name, displayname):
		# Callback for successful login
		self.db._setContext(ctxid)
		self.name = name
		self.displayname = displayname
		self.show()


	@QtCore.pyqtSlot(int)
	def begin_session(self, session):
		# Callback for session begin
		if session > -1:
			self.names[emdash.config.get('session_rectype')] = session

		if self.name:
			self.ui.label_user.setText("%s (%s)"%(self.displayname, self.name))

		self.update_ui()
		self.show()

		
	@QtCore.pyqtSlot()	
	def end_session(self, restart=True):
		# End session
		if self.end:
			return
		# s = self.names.get(emdash.config.get('session_rectype'))
		# self.end = True
		self.close()
		# t = DWarning(parent=self, displayname=self.displayname, session=s)
		# self.controller.finished.connect(t.accept)
		# t.show()
	
	
			
	################
	# Clock
	################
	
	@QtCore.pyqtSlot(str)	
	def tock(self, t):
		# Update session clock
		target = self.names.get(emdash.config.get('session_rectype'))
		if t == None:
			t = ""
		elif target > -1:
			t = """<a href="%s/record/%s/?ctxid=%s">%s</a>"""%(emdash.config.get('host'), target, self.db._ctx.ctxid, t)	
		else:
			t = """<a href="%s?ctxid=%s">%s</a>"""%(emdash.config.get('host'), self.db._ctx.ctxid, t)	
			
		self.ui.label_session.setText(t)



	#####################
	# Event handlers
	#####################
	
	def closeEvent(self, event):
		# event.ignore()
		self.controller.end_session()
	
		
	#####################
	# Comments
	#####################

	@QtCore.pyqtSlot(unicode)
	def addcomment(self, rectype, name=None):
		# Add a comment to a known record
		comment = self.get_comment_text()
		if rectype:
			name = self.names.get(rectype, -1)
		if name < 0:
			error("Could not add comment because there was no %s record yet for this session."%rectype)
			return
		self.signal_set.emit(name, "comments", comment)


	def get_comment_text(self):
		# Get comment text and do any UI updates to reflect this, e.g. clear text field..
		return ""



	#####################
	# Signals to FS Poll Thread
	#####################
	
	@QtCore.pyqtSlot()
	def setpath(self):
		# Select directory to watch for queue
		d = QtGui.QFileDialog(self)
		# d.setDirectory("/")
		d.setFileMode(QtGui.QFileDialog.Directory)
		d.fileSelected.connect(self.controller.fspollthread.setpath, type=QtCore.Qt.QueuedConnection)
		d.fileSelected.connect(self.update_path)		
		d.exec_()
		
		
		
	@QtCore.pyqtSlot()
	def addfiles(self):
		# Directly add some files to the queue
		pass
		# d = QtGui.QFileDialog(self)
		# d.fileSelected.connect(self.controller.fspollthread.setpath, type=QtCore.Qt.QueuedConnection)
		# d.exec_()
			
			
	#####################
	# Misc.
	#####################
			
	def getlabel(self, rectypes):
		"""return the name/recname link for the first rectype found"""
		if not hasattr(rectypes, "__iter__"):
			rectypes = [rectypes]
			
		for r in rectypes:
			if self.names.get(r) != None:
				return """<a href="%s/record/%s?ctxid=%s">%s</a>"""%(emdash.config.get('host'), self.names.get(r), self.db._ctx.ctxid, self.recnames.get(str(self.names.get(r)), r))
		return ""


	def connect_slider(self, i, factor=1000.0):
		slider = getattr(self.ui, 'slider_%s'%i)
		edit = getattr(self.ui, 'edit_%s'%i)
		slider.sliderMoved.connect(functools.partial(self._slider_to_edit, target=edit, factor=factor), type=QtCore.Qt.QueuedConnection)
		edit.textChanged.connect(functools.partial(self._edit_to_slider, target=slider, factor=factor), type=QtCore.Qt.QueuedConnection)
		edit.textChanged.connect(functools.partial(self.set_setting, param=i), type=QtCore.Qt.QueuedConnection)		

	def _slider_to_edit(self, i, target=None, factor=1.0):
		try: i = i / factor
		except: i = 0
		target.setText(str(i))

	def _edit_to_slider(self, i, target=None, factor=1.0):
		try: i = float(i) * factor
		except: i = 0
		target.setValue(int(float(i or 0)))	



	#####################
	# Default wizards..
	#####################

	def _select_grid_wizard(self):
		# Edit the microscopy session we just created to fill in details
		wizard = emdash.emwizard.SelectGridWizard(parent=self, selected=self.names, db=emdash.config.copydb(self.db))
		wizard.signal_done.connect(self.controller.newgrid, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()


	def _name_grid_wizard(self):
		# d, ok = QtGui.QInputDialog.getInt(self, "Existing Grid Imaging", "Record ID")
		d = QtGui.QInputDialog(self)
		d.setLabelText("Record ID")
		d.setIntMaximum(2147483647)
		d.setIntMinimum(0)
		d.setInputMode(QtGui.QInputDialog.IntInput)
		d.intValueSelected.connect(self.controller.selectgrid)
		d.exec_()




class MicroscopeUpload(BaseUpload):
	
	def init(self):
		# Comments
		self.ui.button_addcomment.addAction(QtGui.QAction("... this grid", self, triggered=functools.partial(self.addcomment, rectype='grid_imaging')))
		self.ui.button_addcomment.addAction(QtGui.QAction("... this microscopy session", self, triggered=functools.partial(self.addcomment, rectype='microscopy')))
		self.ui.button_addcomment.addAction(QtGui.QAction("... this microscope", self, triggered=functools.partial(self.addcomment, rectype='microscope')))			

		# This Uploader needs to collect additional data from the user before a session can be started
		self.controller.signal_login.connect(self._newmicroscopy_wizard, type=QtCore.Qt.QueuedConnection)

		# More Drop-down buttons
		self.ui.button_grid.addAction(QtGui.QAction("New Grid Imaging Session", self, triggered=self._newgrid_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("New Freezing Session", self, triggered=self._newfreezing_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("Edit Grid Imaging Session Details", self, triggered=self._edit_grid_imaging))
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session", self, triggered=self._select_grid_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session by Record ID", self, triggered=self._name_grid_wizard))

		self.ui.button_session.addAction(QtGui.QAction("Edit Session", self, triggered=self._edit_microscopy))
			

	def get_ui(self):
		u = emdash.ui.Ui_MicroscopeUpload.Ui_Upload()
		u.setupUi(self)
		self.ui = u
		self._toggle_sessiontype()
		return u	


	def get_controller(self):
		return MicroscopeController(db=emdash.config.copydb(self.db))
			
				
	def update_setting(self, param, value):
		try:
			if param == "id_micrograph":
				self.ui.edit_id_micrograph.setText(value)
			elif param == "ctf_defocus_set":
				pass
			elif param == "tem_magnification_set":
				pass
		except:
			pass


	def update_ui(self, comments=False):
		# Update the UI to reflect these selected records
		self.ui.label_grid.setText(self.getlabel(["grid_imaging"]))		
		self.ui.label_project.setText(self.getlabel(["project","subproject"]))
		self.ui.label_freeze.setText(self.getlabel(["grid_preparation"]))
		
		film_type = self.recs.get(self.names.get(emdash.config.get('session_rectype')), dict()).get('film_type', '')
		if film_type.lower() in emdash.config.get('film_types'):
			self._toggle_sessiontype(session='film')

		self._update_controls()
		if comments:
			self._update_ui_comments()


	def _update_ui_comments(self):
		html = []	
		for c in sorted(self.comments, key=lambda x:x[2], reverse=True):
			t = """<a href="%s/record/%s/?ctxid=%s">%s</a>: %s @ %s: <br /> %s <br /><br />"""%(emdash.config.get('host'), c[0], self.db._ctx.ctxid, self.recnames.get(str(c[0]), c[0]), c[1], c[2], c[3])
			html.append(t)
				
		self.ui.browser_comments.setHtml("".join(html))


	def get_comment_text(self):
		text = self.ui.edit_addcomment.toPlainText()		
		self.ui.edit_addcomment.clear()
		return text



	#####################
	# Wizards!!
	#####################

	@QtCore.pyqtSlot()
	def _newmicroscopy_wizard(self):
		wizard = emdash.emwizard.NewMicroscopyWizard(parent=self, selected=self.names, db=emdash.config.copydb(self.db))
		wizard.signal_done.connect(self.controller.begin_session, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()


	@QtCore.pyqtSlot()
	def _newfreezing_wizard(self):
		wizard = emdash.emwizard.NewFreezingWizard(parent=self, selected=self.names, db=emdash.config.copydb(self.db))
		wizard.signal_done.connect(self.controller.newfreezing, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()



	@QtCore.pyqtSlot()
	def _newgrid_wizard(self):
		wizard = emdash.emwizard.NewGridWizard(parent=self, selected=self.names, db=emdash.config.copydb(self.db))
		wizard.signal_done.connect(self.controller.newgrid, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()


	def _existingmicroscopy_wizard(self):
		# d, ok = QtGui.QInputDialog.getInt(self, "Existing Grid Imaging", "Record ID")
		d = QtGui.QInputDialog(self)
		d.setLabelText("Record ID")
		d.setIntMaximum(2147483647)
		d.setIntMinimum(0)
		d.setInputMode(QtGui.QInputDialog.IntInput)
		d.intValueSelected.connect(self.controller.selectmicroscopy)
		d.exec_()


	def _edit_microscopy(self):
		# Edit the microscopy session we just created to fill in details
		wizard = emdash.emwizard.EditMicroscopyWizard(parent=self, selected=self.names, db=emdash.config.copydb(self.db))
		wizard.signal_done.connect(self.controller.edit_records, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()		


	def _edit_grid_imaging(self):
		if self.names.get('grid_imaging') != None:
			wizard = emdash.emwizard.EditGridImagingWizard(parent=self, selected=self.names, db=emdash.config.copydb(self.db))
			wizard.signal_done.connect(self.controller.edit_records, type=QtCore.Qt.QueuedConnection)
			wizard.exec_()



	##############################
	# Control handlers...
	#############################

	def connect_edits(self):		
		# New Micrograph
		self.ui.button_newmicrograph.clicked.connect(self.controller.newmicrograph, type=QtCore.Qt.QueuedConnection)			

		# Editables
		self.connect_slider('tem_magnification_set', factor=1.)
		self.connect_slider('ctf_defocus_set')
		self.connect_slider('tem_dose_rate')
		self.connect_slider('time_exposure_tem')
		
		self.ui.edit_id_micrograph.textChanged.connect(functools.partial(self.set_setting, param='id_micrograph'), type=QtCore.Qt.QueuedConnection)
		self.ui.button_assess_ice_comments.clicked.connect(self._assess_ice_comments)
		self.ui.button_assess_ice_thick.clicked.connect(self._assess_ice_thick)



	def _toggle_sessiontype(self, session=None):
		if session != 'film':
			self.ui.edit_id_micrograph.hide()
			self.ui.label_newmicrograph.hide()
			self.ui.button_newmicrograph.hide()		
		else:
			self.ui.edit_id_micrograph.show()
			self.ui.label_newmicrograph.show()
			self.ui.button_newmicrograph.show()			



	def _update_controls(self):
		gir = self.recs.get(self.names.get("grid_imaging"), dict())
		if gir.get('tem_magnification_set') != None:
			self.ui.edit_tem_magnification_set.setText(str(gir.get('tem_magnification_set')))
		if gir.get('ctf_defocus_set') != None:
			self.ui.edit_ctf_defocus_set.setText(str(gir.get('ctf_defocus_set')))
		if gir.get('time_exposure_tem') != None:
			self.ui.edit_time_exposure_tem.setText(str(gir.get('time_exposure_tem')))
		if gir.get('tem_dose_rate') != None:
			self.ui.edit_tem_dose_rate.setText(str(gir.get('tem_dose_rate')))


		
	################
	# Comments about ice quality
	################
	
	def _assess_ice_thick(self):
		value = unicode(self.ui.assess_ice_thick.currentText())
		name = self.names.get("grid_imaging", -1)
		if name < 0:
			error("Could not set the ice thickness because you have not yet created a grid_imaging session")
			return
		self.signal_set.emit(name, "assess_ice_thick", value)

		
		
	def _assess_ice_comments(self):
		value = unicode(self.ui.assess_ice_comments.currentText())
		name = self.names.get("grid_imaging", -1)
		if name < 0:
			error("Could not set the ice conditions because you have not yet created a grid_imaging session")
			return
		self.signal_set.emit(name, "assess_ice_comments", value)






# This handler is basically the same as the Microscopy handler, but simpler in a few places.
class ScanUpload(BaseUpload):
	def get_ui(self):
		u = emdash.ui.Ui_ScanUpload.Ui_Upload()
		u.setupUi(self)
		return u		
		

	##############################
	# Control handlers...
	#############################

	def connect_edits(self):		
		# Editables
		self.connect_slider('scan_step')
		self.connect_slider('nikon_gain')
		self.connect_slider('scan_average', factor=10.0)
		self.connect_slider('angstroms_per_pixel')		

		self.ui.edit_scanner_film.editTextChanged.connect(functools.partial(self.set_setting, param='scanner_film'), type=QtCore.Qt.QueuedConnection)
		self.ui.edit_scanner_cartridge.editTextChanged.connect(functools.partial(self.set_setting, param='scanner_cartridge'), type=QtCore.Qt.QueuedConnection)
		



##############################
# Some dialogs: Warning, Login, ...
##############################


class DWarning(QtGui.QDialog):
	
	def __init__(self, displayname=None, session=None, parent=None):
		QtGui.QDialog.__init__(self, parent=parent)		
		self.ui = emdash.ui.Ui_Warning.Ui_Warning()
		self.ui.setupUi(self)
		self.ui.label_user.setText(unicode(displayname))
		self.ui.label_session.setText("""<a href="%s/record/%s/">%s</a>"""%(emdash.config.get('host'), session, session))

	def closeEvent(self, event):
		event.ignore()




class Login(QtGui.QDialog):
	
	signal_getlogin = QtCore.pyqtSignal(unicode, unicode)	
	
	def __init__(self, name=None, password=None, authmsg=None, microscope=None, parent=None):
		QtGui.QDialog.__init__(self, parent=parent)		
		self.ui = emdash.ui.Ui_Login.Ui_Login()
		self.ui.setupUi(self)

		if authmsg:
			self.ui.label_login.setText(authmsg)
		else:
			self.ui.label_login.setText('Login')			

		if name:
			self.ui.edit_name.setText(name)

		if password:
			self.ui.edit_password.setText(password)

		if microscope:
			self.ui.edit_microscope.setText(str(microscope))

		self.ui.edit_host.setText(emdash.config.get('host'))
		

	# turn on before deployment..
	def reject(self):
		return


	def accept(self):
		
		def int_or_none(value):
			if value:
				return int(value)
			return None
		
		if not self.ui.edit_name.text():
			return
		
		host = self.ui.edit_host.text()
		emdash.config.set('host', unicode(host))

		microscope = self.ui.edit_microscope.text()
		emdash.config.set('microscope', int_or_none(microscope))

		if not emdash.config.get('host'):
			msg = "No EMEN2 server specified!"
			error(msg)
			return

		if not emdash.config.get('microscope'):
			msg = "No microscope specified!"
			error(msg)
			return

		self.signal_getlogin.emit(unicode(self.ui.edit_name.text()), unicode(self.ui.edit_password.text()))
		self.done(0)






if __name__ == '__main__':
	main()




__version__ = "$Revision: 1.7 $".split(":")[1][:-1].strip()
