import collections
import copy
import glob
import math
import os
import struct
import subprocess
import tempfile
import time

# XML Etree -- Generally available on Python 2.5+
try:
	import xml.etree.ElementTree as ET
except ImportError:
	ET = None


import emdash.config
import emdash.transport


##################################
# Base FileHandler
##################################

class FileHandler(object):

	request_params = []

	def __init__(self, db=None, name=None, filename=None, filedata=None, rectype=None, applyparams=None, pos=None, options=None, **kwargs):
		"""Provide special support for different filetypes
		@keyparam db DBProxy
		@keyparam name Record ID for download or upload
		@keyparam filename Filename for upload
		@keyparam rectype Record type for default-setting record creation
		@keyparam applyparams Parameters to overlay on new records (see: CCDHandler)
		@keyparam pos Display this FileHandler's position in the overall queue, tuple (position, total)
		@keyparam options Options. See UploadController and DownloadController OptionParsers for details.
		"""

		self.db = db

		# target record name for upload or download
		self.name = name

		# filename for upload..
		self.filename = filename
		self.filedata = filedata
		self.bdos = []

		# new records to apply to uploads
		self.rectype = rectype
		self.applyparams = applyparams or {}

		# default and passed options
		self.options = {
			"recurse": -1,
			"overwrite": False,
			"rename": False,
			"compress": True,
		}

		if options:
			self.options.update(options)

		if kwargs:
			self.options.update(kwargs)

		# internal attributes
		self.tmpfiles = []
		self.checked = False
		self.quiet = False
		self.pos = pos
		self.progress = 0.0
		self.listeners = []
		self.uploaded = {}

		self.init()


	def init(self):
		pass



	######################
	# Upload
	######################

	def upload(self):
		"""Start the upload process for this handler"""

		ret = collections.defaultdict(list)
		recmap = {}

		self.report(pos=self.pos, f=self.filename)
		self.report("Preparing for upload", i=1)

		self.prepare_upload()
		files = self.get_upload_items()

		self.report("Checking and committing records", i=1)

		# Check records, call newrecord to properly init, validate, set permissions, etc.
		# If a particular record doesn't validate, will throw exception
		for f in files:
			if f.newrecord:
				f.newrecord = self._newrecordinit(newrecord=f.newrecord)
								

		# Upload other files/newrecords
		filecount = len(files)
		for count, f in enumerate(files):
			f.setdb(self.db)
			f.setreport(self.report)
			f.setrecmap(recmap)
			
			target_name = f.name

			if filecount > 1:
				f.pos = (count, filecount)

			f.action()
			recmap[target_name] = f.name
			ret[f.filename] = f.bdo

		self.cleanup()
		self.uploaded = ret


	def get_upload_target(self):
		# I don't like this...
		t = None
		if self.uploaded:
			t = self.uploaded.values()[-1].get('record')
		return t
		

	def get_upload_items(self):
		"""Returns records and files to upload"""
		f = emdash.transport.UploadTransport(name=self.name, filename=self.filename)
		return [f]


	def prepare_upload(self):
		"""Do any required prep work before files can be uploaded"""
		self.check_db()


	def _newrecordinit(self, newrecord):
		"""Initialize and validate a new record"""

		tmprec = self.db.newrecord(newrecord.get('rectype'), self.name)
		tmprec.update(newrecord)
		self.db.validaterecord(tmprec)
		return tmprec



	######################
	# Download
	######################

	def download(self):
		"""Start the download process for this handler"""

		overwrite = self.options.get("overwrite")
		rename = self.options.get("rename")
		compress = self.options.get("compress")
		sidecar = self.options.get("sidecar")

		self.report(pos=self.pos, f=self.name)
		self.report("Checking for items to download", i=1)

		recs, bdos = self.get_download_items()

		bdolen = len(bdos)

		if bdos:
			self.report("Found %s items in %s records"%(bdolen, len(recs)), i=1)
		else:
			self.report("Found no downloadable items in %s records"%len(recs), i=1)
			return

		ret = {}

		for count, bdo in enumerate(bdos):
			u = emdash.transport.DownloadTransport(db=self.db, bdo=bdo, report=self.report, compress=self.options.get("compress"), pos=(count, bdolen))
			u.action(overwrite=overwrite, rename=rename, sidecar=sidecar)
			ret[u.filename] = u.bdo

		return ret


	def get_download_items(self):
		"""Find child recods of self.name and get their BDOs"""

		#recs = self.db.getchildren(self.name, "record", self.options.get("recurse"), None, True, True)
		recs = self.db.getchildren(self.name, self.options.get("recurse"))
		recs.append(self.name)
		bdos = self.db.findbinary(record=map(int, recs))
		return recs, bdos



	########################
	# Status updates
	########################

	def addlistener(self, callback):
		self.listeners.append(callback)


	def report(self, msg=None, progress=None, f=None, pos=None, i=0):
		for listener in self.listeners:
			listener(msg=msg, progress=progress, f=f)
		if self.quiet:
			return
		# Normally suppress this...
		if progress != None and msg == None:
			return
		if not msg:
			if pos:
				print "\n%s%s of %s: %s"%("\t"*i, pos[0]+1, pos[1], f)
			else:
				print "\n%s%s"%("\t"*i, f)
		else:
			print "%s%s"%("\t"*i, msg)


	########################
	# Utility methods
	########################


	def _shortsleep(self):
		time.sleep(10)


	def check_db(self):
		pass


	########################
	# Overrideable methods for upload control
	########################

	def cleanup(self):
		for i in self.tmpfiles:
			try:
				self.report("Cleanup temp file: %s"%i, i=1)
				os.unlink(i)
			except Exception, e:
				self.report("Warning: Could not remove %s: %s"%(i, e), i=1)






#################################
# Filetype-specific handlers
##################################


class NewRecordFileHandler(FileHandler):
	def get_upload_items(self):
		newrecord = {}
		newrecord["name"] = -1
		newrecord["rectype"] = self.rectype
		newrecord["parents"] = [self.name]
		return [emdash.transport.UploadTransport(newrecord=newrecord, filename=self.filename, param="file_binary_image")]



class VolumeHandler(NewRecordFileHandler):
	def get_upload_items(self):
		newrecord = {}
		newrecord["name"] = -1
		newrecord["title_structure"] = self.filename
		newrecord["rectype"] = self.rectype
		newrecord["parents"] = [self.name]
		return [emdash.transport.UploadTransport(newrecord=newrecord, filename=self.filename, param="file_binary_image")]



class BoxHandler(FileHandler):
	def get_upload_items(self):
		newrecord = {}
		newrecord["name"] = -1
		newrecord["rectype"] = "box"
		newrecord["parents"] = [self.name]
		newrecord.update(self.applyparams)
		return [emdash.transport.UploadTransport(newrecord=newrecord, filename=self.filename, filedata=self.filedata, compress=False, param="box")]



class ParticleHandler(FileHandler):
	def get_upload_items(self):
		newrecord = {}
		newrecord["name"] = -1
		newrecord["rectype"] = "particles"
		newrecord["parents"] = [self.name]
		return [emdash.transport.UploadTransport(newrecord=newrecord, filename=self.filename, compress=False, param="ptcl")]



class GridImagingFileHandler(FileHandler):
	def check_db(self):
		gi = self.db.getrecord(self.name)

		if gi.get("rectype") != "grid_imaging":
			raise Exception, "This action may only be used with grid_imaging sessions!"

		microscopy = self.db.getparents(self.name, 1, ["microscopy"])
		if not microscopy:
			self.report("\tWARNING! No microscopy record present for grid_imaging session!")
			#raise Exception, "No microscopy record present for grid_imaging session!"



class MicrographHandler(GridImagingFileHandler):
	"""Creates a placeholder for a scan to be uploaded in the future"""

	def get_upload_items(self):
		newrecord = {}
		newrecord["name"] = -1
		newrecord["rectype"] = "micrograph"
		newrecord["id_micrograph"] = os.path.basename(self.filename)
		newrecord.update(self.applyparams)
		newrecord["parents"] = [self.name]
		return [emdash.transport.NewRecordTransport(newrecord=newrecord, filename=self.filename)]
	


class JADASHandler(GridImagingFileHandler):
	"""Creates a placeholder for a scan to be uploaded in the future"""


	def get_upload_items(self):
		# This is run for a .tif file produced by JADAS. Find the associated .xml files, load them, map as many
		# parameters as possible, and attach the raw xml file.
		jadas_params, foundfiles = self.load_jadas_xml()
		newrecord = {}
		newrecord["name"] = -100
		newrecord["rectype"] = "ccd_jadas"
		newrecord["id_micrograph"] = os.path.basename(self.filename)
		newrecord.update(self.applyparams)
		newrecord.update(jadas_params)
		newrecord["parents"] = [self.name]
		
		files = [emdash.transport.UploadTransport(name=-100, filename=self.filename, param='file_binary_image')]
		files[0].newrecord = newrecord
		
		for i in foundfiles:
			files.append(emdash.transport.UploadTransport(name=-100, filename=i, compress=False))
				
		return files


	def load_jadas_xml(self):
		# find related XML files, according to JADAS naming conventions
		# take off the .tif, because the xml could be either file.tif_metadata.xml or file_metadata.xml
		if not ET:
			raise ImportError, "The ElementTree package (xml.etree.ElementTree) is required"
			
		foundfiles = []
		ret = {}		
		for xmlfile in glob.glob('%s_*.xml'%self.filename) + glob.glob('%s_*.xml'%self.filename.replace('.tif','')):
			print "Attempting to load ", xmlfile
			try:
				e = ET.parse(xmlfile)
				root = e.getroot()
				# There should be a loader for each root tag type, e.g. TemParameter -> map_jadas_TemParameter
				loader = getattr(self, 'map_jadas_%s'%root.tag, None)
				if loader:
					ret.update(loader(root))
					foundfiles.append(xmlfile)
			except Exception, e:
				print "Could not load %s: %s"%(xmlfile, e)

		return ret, foundfiles



	def map_jadas_TemParameter(self, root):
		"""One of these long, ugly, metadata-mapping methods"""
		ret = {}
		# Defocus
		ret['defocus_absdac'] = root.find('Defocus/defocus').get('absDac')
		ret['defocus_realphysval'] = root.find('Defocus/defocus').get('relPhisVal')
		ret['intendeddefocus_valinnm'] = root.find('Defocus/intendedDefocus').get('valInNm')
		d = root.find('Defocus/intendedDefocus').get('valInNm')
		if d != None:
			d = float(d) / 1000.0
			ret['ctf_defocus_set'] = d

		# Eos
		ret['eos_brightdarkmode'] = root.find('Eos/eos').get('brightDarkMode')
		ret['eos_darklevel'] = root.find('Eos/eos').get('darkLevel')
		ret['eos_stiglevel'] = root.find('Eos/eos').get('stigLevel')
		ret['eos_temasidmode'] = root.find('Eos/eos').get('temAsidMode')
		ret['eos_htlevel'] = root.find('Eos/eos').get('htLevel')
		ret['eos_imagingmode'] = root.find('Eos/eos').get('imagingMode')
		ret['eos_magcamindex'] = root.find('Eos/eos').get('magCamIndex')
		ret['eos_spectrummode'] = root.find('Eos/eos').get('spectrumMode')
		ret['eos_illuminationmode'] = root.find('Eos/eos').get('illuminationMode')
		ret['eos_spot'] = root.find('Eos/eos').get('spot')
		ret['eos_alpha'] = root.find('Eos/eos').get('alpha')

		# Lens
		ret['lens_cl1dac'] = root.find('Lens/lens').get('cl1Dac')
		ret['lens_cl2dac'] = root.find('Lens/lens').get('cl2Dac')
		ret['lens_cl3dac'] = root.find('Lens/lens').get('cl3Dac')
		ret['lens_cmdac'] = root.find('Lens/lens').get('cmDac')
		ret['lens_il1dac'] = root.find('Lens/lens').get('il1Dac')
		ret['lens_il2dac'] = root.find('Lens/lens').get('il2Dac')
		ret['lens_il3dac'] = root.find('Lens/lens').get('il3Dac')
		ret['lens_il4dac'] = root.find('Lens/lens').get('il4Dac')
		ret['lens_pl1dac'] = root.find('Lens/lens').get('pl1Dac')
		ret['lens_pl2dac'] = root.find('Lens/lens').get('pl2Dac')
		ret['lens_pl3dac'] = root.find('Lens/lens').get('pl3Dac')
		
		# Def
		ret['def_gunshiftx'] = root.find('Def/def').get('gunShiftX')
		ret['def_gunshifty'] = root.find('Def/def').get('gunShiftY')
		ret['def_guntiltx'] = root.find('Def/def').get('gunTiltX')
		ret['def_guntilty'] = root.find('Def/def').get('gunTiltY')
		ret['def_beamshiftx'] = root.find('Def/def').get('beamShiftX')
		ret['def_beamshifty'] = root.find('Def/def').get('beamShiftY')
		ret['def_beamtiltx'] = root.find('Def/def').get('beamTiltX')
		ret['def_beamtilty'] = root.find('Def/def').get('beamTiltY')			
		ret['def_clstigx'] = root.find('Def/def').get('clStigX')
		ret['def_clstigy'] = root.find('Def/def').get('clStigY')
		ret['def_olstigx'] = root.find('Def/def').get('olStigX')
		ret['def_olstigy'] = root.find('Def/def').get('olStigY')
		ret['def_ilstigx'] = root.find('Def/def').get('ilStigX')
		ret['def_ilstigy'] = root.find('Def/def').get('ilStigY')
		ret['def_imageshiftx'] = root.find('Def/def').get('imageShiftX')
		ret['def_imageshifty'] = root.find('Def/def').get('imageShiftY')
		ret['def_plax'] = root.find('Def/def').get('plaX')
		ret['def_play'] = root.find('Def/def').get('plaY')
		
		# HT
		ret['ht_ht'] = root.find('HT/ht').get('ht')
		ret['ht_energyshift'] = root.find('HT/ht').get('energyShift')
		
		# MDS
		ret['mds_mdsmode'] = root.find('MDS/mds').get('mdsMode')
		ret['mds_blankingdef'] = root.find('MDS/mds').get('blankingDef')
		ret['mds_defx'] = root.find('MDS/mds').get('defX')
		ret['mds_defy'] = root.find('MDS/mds').get('defY')
		ret['mds_blankingtype'] = root.find('MDS/mds').get('blankingType')
		ret['mds_blankingtime'] = root.find('MDS/mds').get('blankingTime')
		ret['mds_shutterdelay'] = root.find('MDS/mds').get('shutterDelay')
		
		# Photo
		ret['photo_exposuremode'] = root.find('PHOTO/photo').get('exposureMode')
		ret['photo_manualexptime'] = root.find('PHOTO/photo').get('manualExpTime')
		ret['photo_filmtext'] = root.find('PHOTO/photo').get('filmText')
		ret['photo_filmnumber'] = root.find('PHOTO/photo').get('filmNumber')
		
		# GonioPos
		ret['goniopos_x'] = root.find('GonioPos/gonioPos').get('x')
		ret['goniopos_y'] = root.find('GonioPos/gonioPos').get('y')
		ret['goniopos_z'] = root.find('GonioPos/gonioPos').get('z')
		ret['goniopos_tiltx'] = root.find('GonioPos/gonioPos').get('tiltX')
		ret['goniopos_rotortilty'] = root.find('GonioPos/gonioPos').get('rotOrTiltY')

		return ret
		
		
		
	def map_jadas_DigitalCameraParameter(self, root):
		attrmap = {
			'CameraName': 'ccd_id',
			'AreaTop': 'digicamprm_areatop',
			'AreaBottom': 'digicamprm_areabottom',
			'AreaLeft': 'digicamprm_arealeft',
			'AreaRight': 'digicamprm_arearight',
			'Exposure': 'time_exposure_tem',
			'Binning': 'binning',
			'PreIrradiation': 'digicamcond_preirradiation',
			'BlankingTime': 'digicamcond_blankingtime',
			'BlankBeam': 'digicamcond_blankbeam',
			'CloseScreen': 'digicamcond_closescreen',
			'DataFormat': 'digicamcond_dataformat'		
		}

		ret = {}
		for i in root.findall('*/tagCamPrm'):
			param = attrmap.get(i.get('tagAttrName'))
			value = i.get('tagAttrVal')
			if param != None and value != None:
				ret[param] = value		

		return ret
		
		
		
	def map_jadas_IntensityBasedHoleSelection(self, root):
		ret = {}
		return ret
		
			


class CCDHandler(GridImagingFileHandler):

	request_params = [
		'tem_magnification_set',
		'ctf_defocus_set',
		'angstroms_per_pixel',
		'time_exposure_tem',
		'tem_dose_rate',
		'current_screen',
		'ice_type',
		'assess_ice_thick',
		'assess_ice_comments',
		'assess_image_quality',
		'ccd_id',
		'status_energy_filter',
		'binning'
		]


	def get_upload_items(self):
		newrecord = {}
		newrecord["name"] = -1
		newrecord["rectype"] = "ccd"
		newrecord["id_ccd_frame"] = os.path.basename(self.filename)
		newrecord.update(self.applyparams)
		newrecord["parents"] = [self.name]
		# read metadata
		return [emdash.transport.UploadTransport(newrecord=newrecord, filename=self.filename, param="file_binary_image")]




class ScanHandler(GridImagingFileHandler):

	request_params = [
		'scanner_film',
		'scanner_cartridge',
		'scan_average',
		'nikon_gain',
		'scan_step',
		'angstroms_per_pixel'
		]


	def get_upload_items(self):
		print "\tChecking for existing micrograph..."
		idmap = collections.defaultdict(set)
		mc = self.db.getchildren(self.name, 1, "micrograph")
		mc = self.db.getrecord(mc)
		for rec in mc:
			i = rec.get('id_micrograph', '').strip().lower()
			idmap[i].add(rec.get('name'))


		outfile = self.filename

		# This is an ugly hack until I think of a better way
		opts = {}
		try:
			# tif2mrc, bin, invert, odconversion
			for i in ['tif2mrc', 'bin', 'invert', 'odconversion']:
				opts[i] = emdash.config.get(i)
				# getattr(emdash.config, i, None)
			print "Using options for ScanHandler:", opts
		except:
			pass

		if opts.get('tif2mrc'):
			outfile = outfile.replace('.tif', '.mrc')
			args = []
			# if python:
			# 	args.append(python)
			args.append('nikontiff2mrc.py')
			if opts.get('bin') != None:
				args.append('--bin=%s'%opts.get('bin'))
			if opts.get('invert') != None:
				args.append('--invert=%s'%opts.get('invert'))
			if opts.get('odconversion') != None:
				args.append('--ODconversion=%s'%opts.get('odconversion'))

			args.append(self.filename)
			args.append(outfile)
			
			print "running: %s"%args
			a = subprocess.Popen(args)
			a.wait()

		# Try to find matches between the current filename and items in the imaging session
		match = os.path.basename(outfile.split(".")[0].strip().lower())
		matches = idmap[match]

		if len(idmap[match]) == 0:
			print "\tCould not find micrograph for %s -- creating new micrograph."%match

			mrec = {}
			mrec["name"] = -1
			mrec["rectype"] = "micrograph"
			mrec["parents"] = [self.name]
			mrec["id_micrograph"] = match

			newrecord = {}
			newrecord["name"] = -2
			newrecord["rectype"] = 'scan'
			newrecord["parents"] = [-1]
			newrecord.update(self.applyparams)

			m = emdash.transport.NewRecordTransport(newrecord=mrec)
			s = emdash.transport.UploadTransport(newrecord=newrecord, filename=outfile, param="file_binary_image")

			sidecar = s.sidecar_read()
			if sidecar:
				print "\tThis scan already appears to be uploaded! Check record ID %s"%sidecar.get('name')
				return []

			return [m, s]


		elif len(idmap[match]) == 1:
			matches = matches.pop()
			print "\tFound match for %s: %s"%(match, matches)
			newrecord = {}
			newrecord["name"] = -1
			newrecord["rectype"] = 'scan'
			newrecord["parents"] = [matches]
			newrecord.update(self.applyparams)

			return [emdash.transport.UploadTransport(newrecord=newrecord, filename=outfile, param="file_binary_image")]


		elif len(idmap[match]) > 1:
			print "\tAmbiguous matches for %s: %s -- skipping"%(match, matches)
			return []





class StackHandler(GridImagingFileHandler):

	# See notes at bottom of file
	header_labels = [
		# X,Y,Z size
		['stack_size_nx', 'i', 0],
		['stack_size_ny', 'i', 0],
		['stack_size_nz', 'i', 0],
		## Mode: mode, 0 = unsigned byte, 1 = short int, 2 = float, 3 = short*2, 4 = float*2
		['stack_data_mode', 'i', 0],
		## Starting point of sub image: nxstart, nystart, nzstart
		['stack_start_nx', 'i', 0],
		['stack_start_ny', 'i', 0],
		['stack_start_nz', 'i', 0],
		## Grid Size: mx, my, mz
		['stack_size_mx', 'i', 0],
		['stack_size_my', 'i', 0],
		['stack_size_mz', 'i', 0],
		## Cell size: xlen, ylen, zlen... pixel spacing = xlen/mx
		['stack_size_xlen', 'i', 0],
		['stack_size_ylen', 'i', 0],
		['stack_size_zlen', 'i', 0],
		## Cell angles: alpha, beta, gamma
		['stack_angle_alpha', 'f', 0.0],
		['stack_angle_beta', 'f', 0.0],
		['stack_angle_gamma', 'f', 0.0],
		## Map column. Ignored by IMOD: mapc, mapr, maps. (1=x, 2=y, 3=z)
		['stack_map_mapc', 'i', 0],
		['stack_map_mapr', 'i', 0],
		['stack_map_maps', 'i', 0],
		## Minimum pixel value, amin
		['stack_pixel_min', 'f', 0.0],
		## Maximum pixel value, amax
		['stack_pixel_max', 'f', 0.0],
		## Mean pixel value, amean
		['stack_pixel_mean', 'f', 0.0],
		## Image type, ispg
		['stack_data_ispg', 'h', 0],
		## Space group number, nsymbt
		['stack_data_nsymbt', 'h', 0],
		## Number of bytes in extended header, next
		['stack_data_extheadersize', 'i', 0],
		## Creator ID, creatid
		['stack_data_creatid', 'h', 0],
		# 30 bytes of unused data: See explanation above
		['stack_unused', '30s', ''],
		## Number of bytes per section (SerialEM format), nint
		['stack_data_bytespersection', 'h', 0],
		## Flags for which types of short data (SerialEM format), nreal. (See above documentation.)
		['stack_data_extheaderflags', 'h', 0],
		# 28 bytes of unused data: See explanation above
		['stack_unused2', '28s', ''],
		['stack_data_idtype', 'h', 0],
		['stack_data_lens', 'h', 0],
		['stack_data_nd1', 'h', 0],
		['stack_data_nd2', 'h', 0],
		['stack_data_vd1', 'h', 0],
		['stack_data_vd2', 'h', 0],
		['stack_data_tiltangles_orig', 'fff', [0.0, 0.0, 0.0]],
		['stack_data_tiltangles_current', 'fff', [0.0, 0.0, 0.0]],
		['stack_data_xorg', 'f', 0.0],
		['stack_data_yorg', 'f', 0.0],
		['stack_data_zorg', 'f', 0.0],
		['stack_data_cmap', '4s', ''],
		# big/little endian flag
		['stack_data_stamp', '4s', ''],
		['stack_data_rms', 'f', 0.0],
		['stack_data_nlabl', 'i', 0],
		# spliced seperately: see after
		['stack_data_labels', '80s'*10, '']
	]


	extheader_flags = {
		1: {
			'pack': 'h',
			'load': lambda x:[x[0] / 100.0],
			'dump': lambda x:[x[0] * 100],
			'dest': ['specimen_tilt'],
			},

		2: {
			'pack': 'hhh',
			'load': lambda x:[x],
			'dump': lambda x:[x],
			'dest': ['stack_data_montage']
			},

		4: {
			'pack': 'hh',
			'load': lambda x:[x[0] / 25.0 , x[1] / 25.0],
			'dump': lambda x:[x[0] * 25   , x[1] * 25],
			'dest': ['stack_stagepos_x', 'stack_stagepos_y']
			},

		8: {
			'pack': 'h',
			'load': lambda x:[x[0] / 10.0],
			'dump': lambda x:[x[0] * 10],
			'dest': ['tem_magnification_set']
			},

		16: {
			'pack': 'h',
			'load': lambda x:[x[0] / 25000.0],
			'dump': lambda x:[x[0] * 25000],
			'dest': ['stack_intensity']
			},

		32: {
			'pack': 'hh',
			'load': lambda x:[sign(x[0])*(math.fabs(x[0])*256+(math.fabs(x[1])%256))*2**(sign(x[1])*(int(math.fabs(x[1]))/256))],
			'dump': lambda x:[0,0],
			'dest': ['stack_dose']
			}
	}


	def init(self):
		pass


	def prepare_upload(self):
		try:
			self.header = self.get_header()
			self.slices = self.get_extheader()
			self.update_maxangles()
			self.header['stack_filename'] = self.filename
		except Exception, inst:
			raise ValueError, "This does not appear to be a SerielEM stack: %s"%inst

		# Extract headers for backup, unstack file and compress, add to tmpfiles to be removed after upload
		self.check_db()

		# self.tmpdir = tempfile.mkdtemp()
		# self.raw_headers = self.backup_headers()
		# self.tmpfiles.extend(self.raw_headers)
		# self.unstacked_files = self.unstack()
		# self.tmpfiles.extend(self.unstacked_files)


	def get_upload_items(self):
		"""Get basic records and files to put in database. These will be filled out more fully with db.newrecord, etc."""

		stackrec = copy.copy(self.header)
		stackrec["name"] = -1
		stackrec["parents"] = [self.name]
		return [emdash.transport.UploadTransport(newrecord=stackrec, filename=self.filename, param="file_binary_image")]
		
		# files = [emdash.transport.UploadTransport(name=-100, filename=i, compress=False) for i in self.raw_headers]
		# files[0].newrecord = stackrec
		# for offset, (s, s_filename) in enumerate(zip(self.slices, self.unstacked_files)):
		# 	s["name"] = -(offset+101)
		# 	s["parents"] = [-100]
		# 	files.append(emdash.transport.UploadTransport(newrecord=s, filename=s_filename, param="file_binary_image"))


	########
	# Stack specific methods, mostly header related

	def update_maxangles(self):
		if self.slices[0].has_key("specimen_tilt"):
			tilts = [i['specimen_tilt'] for i in self.slices]

		self.header["stack_maxangle"] = max(tilts)
		self.header["stack_minangle"] = min(tilts)


	def get_header(self):
		"""Read an IMOD/SerialEM/MRC stack header"""
		f = open(self.filename,"rb")
		h = f.read(1024)
		f.close()
		return self._handle_header(h)


	def get_extheader(self):
		if not self.header:
			self.header = self.get_header()

		if not self.header["stack_data_extheadersize"]:
			return []

		f = open(self.filename, "rb")
		h = f.read(1024)
		eh = f.read(self.header["stack_data_extheadersize"])
		f.close()

		return self._handle_extheader(eh)


	def _get_offset_and_size(self):
		# 4    int     mode;      Types of pixel in image.  Values used by IMOD:
		# 	 0 = unsigned bytes,
		# 	 1 = signed short integers (16 bits),
		# 	 2 = float,
		# 	 3 = short * 2, (used for complex data)
		# 	 4 = float * 2, (used for complex data)
		#                          6 = unsigned 16-bit integers (non-standard)
		# 	16 = unsigned char * 3 (for rgb data, non-standard)

		header_offset = 1024 + self.header["stack_data_extheadersize"]
		depth = 2
		dmode = self.header['stack_data_mode']
		if dmode in [0]:
			depth = 1
		elif dmode in [2, 3]:
			depth = 4
		elif dmode in [4]:
			depth = 8
		elif dmode in [16]:
			depth = 3

		slicesize = self.header["stack_size_nx"] * self.header["stack_size_ny"] * depth

		return header_offset, slicesize


	def _handle_header(self, h):
		"""Extract data from header string (1024 bytes) and process"""

		d={}
		offset = 0
		for dest, format, default in self.header_labels:
			size = struct.calcsize(format)
			value = struct.unpack(format, h[offset:offset+size])
			if len(value) == 1:
				d[dest] = value[0]
			else:
				d[dest] = value
			offset += size

		# Process data labels
		n = d["stack_data_nlabl"]

		d["stack_data_labels"] = [i.strip() for i in d["stack_data_labels"][:n]]+[""]*(10-n)

		# Include rectype
		d["rectype"] = "stack"
		d["name"] = -1

		# This guy gives us trouble
		d['stack_data_stamp'] = d['stack_data_stamp'].partition('\x00')[0]

		# Delete unused items
		try:
			del d["stack_unused"]
			del d["stack_unused2"]
		except:
			pass
		
		return d


	def _handle_extheader(self, eh):
		"""Process extended header"""

		d = self.header
		ed = []
		keys = self._extheader_getkeys(d["stack_data_extheaderflags"])

		print d['stack_size_nz']

		offset = 0
		for i in range(0, d["stack_size_nz"]):
			sslice = {}
			for key in keys:
				_k = self.extheader_flags.get(key)
				size = struct.calcsize(_k['pack'])
				# print "Consuming %s bytes (%s:%s) for %s"%(size, i+offset, i+offset+size, _k['dest'])
				value = struct.unpack(_k['pack'], eh[offset: offset+size])
				value = _k['load'](value)

				for _i in zip(_k['dest'], value):
					sslice[_i[0]] = _i[1]

				offset += size

			sslice["rectype"] = "stackimage"

			ed.append(sslice)

		return ed


	def _extheader_getkeys(self, flags):
		keys = []
		for i in sorted(self.extheader_flags.keys()):
			if flags & i:
				keys.append(i)
		return keys


	
	# ian: This was just too complicated and fragile. I've decided the policy will always
	# be to upload the original, 100% untouched file. The server can process it into
	# additional views, or a more advanced tile format for volumes, but the original .st
	# file should be archived as-is.
	
	# def download_postprocess(self, names, ret):
	# 	rec = self.db.getrecord(self.name)
	# 	print "Rebuilding .st file. It is generally a good idea to run this download in a new directory."
	# 
	# 	workfile = rec.get('stack_filename')
	# 	if os.access(workfile, os.F_OK):
	# 		raise ValueError, "File %s already exists -- not going to overwrite. Exiting."
	# 
	# 	f = open(workfile, "wb")
	# 
	# 	print "Making header"
	# 	a = self._make_header(rec)
	# 	f.write(a)
	# 
	# 	print "Making extheader"
	# 	slices = self.db.getrecord(sorted(self.db.getchildren(self.name, 1, "stackimage"), reverse=True))
	# 
	# 
	# 	hl = rec.get('stack_data_extheaderflags')
	# 	hs = rec.get('stack_data_extheadersize')
	# 	b = self._make_extheader(slices, stack_data_extheaderflags=hl)
	# 	padsize = hs - len(b)
	# 	f.write(b)
	# 	print "\t...size was %s, padding by %s to %s / %s"%(len(b), padsize, len(b)+padsize, hs)
	# 	f.write("\0"*padsize)
	# 
	# 	print "Adding images to stack"
	# 
	# 	reti = {}
	# 	for k,v in ret.items():
	# 		reti[v.get("name")] = k
	# 
	# 	for rec in slices:
	# 		fn = reti.get(rec.get("name"))
	# 		#print fn
	# 		fnopen = open(fn, "rb")
	# 		fnopen.seek(1024)
	# 		data = fnopen.read()
	# 		f.write(data)
	# 		# print "\t...wrote %s bytes"%len(data)
	# 		fnopen.close()
	# 		os.unlink(fn)
	# 
	# 	f.close()


	# def backup_headers(self):
	# 	fin = open(self.filename, 'rb')
	# 	raw_header = fin.read(1024)
	# 	raw_extheader = fin.read(self.header["stack_data_extheadersize"])
	# 	fin.close()
	# 
	# 	ret = []
	# 
	# 	header_outfile = "%s/%s.header"%(self.tmpdir, os.path.basename(self.filename))
	# 
	# 	fout = open(header_outfile, 'wb')
	# 	fout.write(raw_header)
	# 	fout.close()
	# 	ret.append(header_outfile)
	# 
	# 	if raw_extheader:
	# 		extheader_outfile = "%s/%s.header"%(self.tmpdir, os.path.basename(self.filename))
	# 		fout = open(extheader_outfile, 'wb')
	# 		fout.write(raw_extheader)
	# 		fout.close()
	# 		ret.append(extheader_outfile)
	# 
	# 	return ret


	# def unstack(self, slices=None):
	# 	"""Unstack slices into new files.
	# 	@keyparam slices Single slice or list of slices to unstack; default is all.
	# 	@return Filenames of unstacked files
	# 	"""
	# 
	# 	raw_header = self._make_unstacked_header()
	# 	header_offset, size = self._get_offset_and_size()
	# 	fstack = open(self.filename, "rb")
	# 
	# 	ret = []
	# 
	# 	if slices and not hasattr(slices, "__iter__"):
	# 		slices = [slices]
	# 	if not slices:
	# 		slices = range(len(self.slices))
	# 
	# 	for n in slices:
	# 		read_offset = header_offset + (size*n)
	# 
	# 		fout_name = "%s/%s.%s.mrc"%(self.tmpdir, os.path.basename(self.filename), n)
	# 
	# 		self.report("Unstacking frame %s into %s"%(n, fout_name), i=1)
	# 
	# 		# check_filenames = [fout_name]
	# 		# check_filenames.append(fout_name+".gz")
	# 		# if not filter(None, map(os.path.exists, check_filenames)):
	# 
	# 		#ret.append(fout_name)
	# 		#continue
	# 
	# 		fout = open(fout_name, "wb")
	# 		fout.write(raw_header)
	# 		fstack.seek(read_offset)
	# 
	# 		blen = 1024*1024*8
	# 		for buffer_size in [blen]*(size/blen)+[size%blen]:
	# 			copy_buffer = fstack.read(buffer_size)
	# 			fout.write(copy_buffer)
	# 
	# 		fout.close()
	# 
	# 		ret.append(fout_name)
	# 
	# 	fstack.close()
	# 	return ret


	# def stack(self, outfile=None):
	# 	pass


	# def _make_unstacked_header(self):
	# 
	# 	h = copy.deepcopy(self.header)
	# 	for i in ['stack_size_nz', 'stack_size_zlen']:
	# 		h[i] = 1
	# 	for i in ['stack_data_extheadersize', 'stack_data_bytespersection', 'stack_data_extheaderflags']:
	# 		h[i] = 0
	# 
	# 	rawheader = self._make_header(h)
	# 	return rawheader


	# def _make_header(self, header):
	# 	rawheader = []
	# 
	# 	header["stack_unused"] = ""
	# 	header["stack_unused2"] = ""
	# 
	# 	for dest, format, default in self.header_labels:
	# 		value = header.get(dest, default)
	# 		size = struct.calcsize(format)
	# 		if hasattr(value,"__iter__"):
	# 			value = struct.pack(format, *value)
	# 		else:
	# 			value = struct.pack(format, value)
	# 		# print dest, format, header.get(dest), value
	# 		rawheader.append(value)
	# 
	# 	return "".join(rawheader)


	# def _make_extheader(self, slices, stack_data_extheaderflags=0):
	# 	eh = []
	# 
	# 	# ian: todo: calculate keys based on info in DB instead of extheaderflags
	# 	keys = self._extheader_getkeys(stack_data_extheaderflags)
	# 
	# 	for s in slices:
	# 		for key in keys:
	# 			_k = self.extheader_flags.get(key)
	# 
	# 			value = [s[i] for i in _k['dest']]
	# 			value = _k['dump'](value)
	# 			if hasattr(value, "__iter__"):
	# 				value = struct.pack(_k['pack'], *value)
	# 			else:
	# 				value = struct.pack(_k['pack'], value)
	# 
	# 			# print key, _k, [s[i] for i in _k['dest']]
	# 			eh.append(value)
	# 
	# 	return "".join(eh)






##################################
# Util
##################################


def sign(a):
	if a>0: return 1
	return -1


map_filename_rectype = {
	".st": "stack",
	".dm3": "ccd",
	".tif": "jadas",
	'.mrc': 'scan'
}

map_rectype_handler = {
	"ccd": CCDHandler,
	"stack": StackHandler,
	"scan": ScanHandler,
	"volume": VolumeHandler,
	"micrograph": MicrographHandler,
	"jadas": JADASHandler,
	"none": FileHandler
}


def gethandler(filename=None, rectype=None):
	if not rectype:
		ext = os.path.splitext(filename)[-1]
		rectype = map_filename_rectype.get(ext)
	# print "For rectype/filename %s %s got handler "%(rec.get('rectype'), filename), handler		
	return map_rectype_handler.get(rectype, FileHandler)





__version__ = "$Revision: 1.6 $".split(":")[1][:-1].strip()
